#include <nes6502.h>
#include <mapper.h>
#include "globals.h"
// #include <stdio.h>
#include "types.h"
#include "cpu-glue.h"
#include "dc-glue.h"

// #define printf if(0) printf

int ESI = 0;

extern void input_shim_m6502(int);
extern void output_shim_m6502(int,int);
extern int nmi_shim_m6502(int);
extern void mmc1_shim_m6502(int,int);

#if 1
void mem_write(uint32 Addr, uint8 Value) 
{
	void ((*fxn)(int,int));
	profile_end(6);
//	printf("Wr6502 at %04x (%02x) (CLOCK %d)\n", Addr, Value, CLOCK);
// if (Addr > 0xffff) { printf("Addr>0xffff\n"); quit(); }
	if (Addr & 0x8000) { 
		/* Redirect addresses $8000-$FFFF through mapper. */
		/* asm linkage addr is stored in Mapper[MAPPERNUM];
		 * from there asm calls C mmc1, mmc3, etc. */
		fxn = Mapper[MAPPERNUMBER];
		/* Can't remember how to cast to void fxn(int,int) ;) */
		fxn(Addr,Value);
	} else if (Addr >= 0x2000 && Addr < 0x6000) {
	    /* Redirect addresses $2000-$5FFF to Output. */
		/* table.x86 redirects are inconsistent. */
		output_shim_m6502(Addr,Value);
	} else {
		/* RAM */
		/* (ZPMEM and NVRAM are contained in RAM, so I'm omitting the
		 * extra tests for convenience.) */
		RAM[Addr] = Value;
	}
	profile_start(6);
}
#endif

uint8 mem_read(uint32 Addr) 
{
	int ret;

// if (Addr > 0xffff) { printf("Addr>0xffff\n"); quit(); }
//   printf("Rd6502 at %04x (CLOCK %d) ->", Addr, CLOCK);
//       printf("MAPTABLE[%d]=%p  ->", Addr>>12,MAPTABLE[Addr>>12]+Addr);
	if (Addr >= 0x2000 && Addr < 0x6000) {
		profile_end(6);
		input_shim_m6502(Addr);
		profile_start(6);
//		printf(" -> %02x\n", INRET & 0xFF);
		return INRET & 0xFF;
	} else {
		ret=*(MAPTABLE[Addr>>12] + Addr); 
//		printf(" %02x\n", ret);
		return ret;
	}
}

/*
INLINE uint8 bank_readbyte(register uint32 Addr)
{ 
// if (Addr > 0xffff) { printf("Addr>0xffff\n"); quit(); }
	return *(MAPTABLE[Addr>>12]+Addr); 
}

INLINE uint32 bank_readaddress(register uint32 Addr) { 
// if (Addr > 0xffff) { printf("Addr>0xffff\n"); quit(); }
	return (bank_readbyte(Addr+1) << 8) | (bank_readbyte(Addr));
}

INLINE void bank_writebyte(register uint32 Addr, register uint8 Value) { 
// if (Addr > 0xffff) { printf("Addr>0xffff\n"); quit(); }
	RAM[Addr] = Value; 
}

INLINE uint32 zp_address(register uint8 Addr) { return (*(MAPTABLE[0]+Addr+1) << 8) | *(MAPTABLE[0]+Addr);
}
*/

#undef printf

#ifndef DYNAMIC_RECOMPILE
void START(void) 
{
	int nmi;

	ram = RAM;  // or MAPTABLE[0]?
	stack_page = ram + STACK_OFFSET; // or STACK?
	nes6502_init();
	nes6502_reset();

	for (;;) {
		profile_start(6);
		ESI += nes6502_execute(100);
		profile_end(6);
//		printf(";\n");

		if (ESI >= 0) {
			nmi = nmi_shim_m6502(reg_PC);
			if (nmi == INT_NMI) {
//				printf("NMI\n");
				nes6502_nmi();   // And nmi executes at top of next _execute.
			} else if (nmi == INT_IRQ) {
				nes6502_irq();
			} else if (nmi == INT_QUIT) {
				break;
			}
			ESI = CTNI;
		}
	}
}
#endif
