/*
 * TuxNES
 *
 * Written by Quor as part of the Nestra codebase
 * Extensively modified by The Mighty Mike Master
 * Also modified by Benjamin C. W. Sittler for TuxNES
 *
 * Description: This file handles the output of the graphics and the input
 * from the keyboard and the joystick.
 *
 * $Id: dc.c,v 1.12 2002/02/28 00:28:48 jim Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#undef USE_TA

#ifndef HAVE_DC
#ifdef HAVE_FEATURES_H
#include <features.h>
#endif /* HAVE_FEATRES_H */

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#else  /* !HAVE_DC */
#include "dc-glue.h"
#include <kos.h>
#endif /* !HAVE_DC */

#include "consts.h"
#include "globals.h"
#include "mapper.h"
#include "renderer.h"
#include "sound.h"

/* imports */
void	quit(void);
void	START(void);
void	fbinit(void);

#ifdef HAVE_DC
#define HAVE_DC_TIME
// Should be able to remove above line, redundant

// FIXME
#undef fprintf
#define fprintf(ARGS...) ;
/* exported functions */
int	InitDisplayDC(int argc, char **argv);
void	UpdateColorsDC(void);
int	UpdateDisplayDC(void);

int	InitDisplayDCtile(int argc, char **argv);
void	UpdateColorsDCtile(void);
int	UpdateDisplayDCtile(void);

/* For tile cache; may prefer external header. */
extern void tile_cache_initialize();
extern void tile_cache_invalidate_all();

static void	InitScreenshotDC(void);
static void	SaveScreenshotDC(void);
static int	HandleMapleDC(void);
static void init_timer(void);
extern int	screenshotnumber;
extern char	*screenshotfile;
extern unsigned char	needsredraw;		/* Refresh screen display */
extern unsigned char	redrawbackground;	/* Redraw tile background */
extern unsigned char	redrawall;		/* Redraw all scanlines */
extern unsigned int	currentbgcolor, oldbgcolor;

uint16 *bufferDC[2] = {0, 0};
int frameno;

uint16 whitepixel, blackpixel;

uint16 paletteDC[64], palette2DC[64];
uint16 basepixel;

/* TuxNES can only use 2 controllers */
#define MAX_CONTROLLERS 2
uint8 maple_controller[MAX_CONTROLLERS];

uint8 maple_nth_device(int code, int n);

#  ifdef USE_TA
uint32 txraddr;
void draw_screen_poly();
#  endif

#endif

#ifdef HAVE_DC

#if 0
void InitScreenshotDC(void) {}
void SaveScreenshotDC(void) {}
#endif

void init_timer(void)
{
	timer_prime(2,1,0);  /* second arg is junk */
	timer_clear(2);

	/* kos doesn't let us set the timer to (uint)-1 */
	TIMER32(TCOR2) = 0xffffffff; /* TCOR2 */
	TIMER32(TCNT2) = 0xffffffff; /* TCNT2 */

	timer_start(2);
	renderer_data.basetime = timer_count(2);
}

/* First with a given function code... */
uint8 maple_nth_device(int code, int n) {
	int port, unit;
	int i = 0;

	for (port=0; port<4; port++) {
		for (unit=0; unit<6; unit++) {
			if (maple_device_func(port, unit) & code) {
				if (i == n)
					return maple_create_addr(port, unit);
				i++;
			}
		}
	}
	return 0;
}

int
HandleMapleDC(void)
{
	// FIXME this means all buttons are depressed
	static unsigned int last_cond[MAX_CONTROLLERS] = { 0 };
	int i;
	cont_cond_t cond;

	fprintf(stderr,"handlemapledc: enter\n");

	for (i = 0; i < MAX_CONTROLLERS; i++) {

		if (!maple_controller[i]) return 0;

		if (cont_get_cond(maple_controller[i], &cond)) {
			fprintf(stderr, "Error getting controller %d status\n", i);
			/* libdream cleanup? */
			return 1;
		}

		/* START+B to quit */
		if (!(cond.buttons & (CONT_START | CONT_B))) {
			fprintf(stderr,"escape sequence pressed, quitting\n");
			return 1;                      
		}

		/* Unisystem stuff not implemented */

		fprintf(stderr,"cond.buttons: %04x\n", cond.buttons);
		controller[i] = 0;
		if (!(cond.buttons & CONT_START)) { controller[i] |= STARTBUTTON; }
		if (!(cond.buttons & CONT_A)) { controller[i] |= BUTTONA; }
		if (!(cond.buttons & CONT_X)) { controller[i] |= BUTTONB; }
		if (!(cond.buttons & CONT_DPAD_LEFT)) { controller[i] |= LEFT; }
		if (!(cond.buttons & CONT_DPAD_RIGHT)) { controller[i] |= RIGHT; }
		if (!(cond.buttons & CONT_DPAD_UP)) { controller[i] |= UP; }
		if (!(cond.buttons & CONT_DPAD_DOWN)) { controller[i] |= DOWN; }
		if (!(cond.buttons & CONT_Y)) { controller[i] |= SELECTBUTTON; }

		if (!(cond.buttons & CONT_B))
			if ((last_cond[i] & CONT_B)) tile_cache_invalidate_all();

		last_cond[i] = cond.buttons;
	}
	fprintf(stderr,"handlemapledc: leave\n");
	return 0;

}

int
InitDisplayDC(int argc, char **argv)
{
  int x;

/*  struct timeval time; */

  if (magstep < 1) {
    magstep = 1;
  }

  if (magstep > maxsize) {
    fprintf (stderr, "Warning: Enlargement factor %d is too large!\n",
	     magstep);
    magstep = maxsize;
  }

	/* ensure hardcode magstep for now */
	magstep = 1;
#ifdef USE_TA
	vid_init(DM_640x480, PM_RGB565);
	vid_empty();
	ta_init();
#else
	if (magstep == 2) 
		vid_init(DM_640x480, PM_RGB565);
	else
		vid_init(DM_320x240, PM_RGB565);
	vid_empty();  /* vid_init only clears the first page */
#endif

      blackpixel = 0x0000;
      whitepixel = 0xffff;
      for (x=0; x < 24; x++)
	  palette[x] = palette2[x] = blackpixel;

      for (x=0; x < 64; x++)
	{
	  /* RGB565 */
	  uint8 r,g,b;
	  r = (NES_palette[x] >> 16 >> 3) & 0x1f;
	  g = (NES_palette[x] >>  8 >> 2) & 0x3f;
	  b = (NES_palette[x] >>       3) & 0x1f;
	  paletteDC[x] = (r << 11) | (g << 5) | (b);
	  palette2DC[x] = blackpixel;
	}

      if (scanlines && (scanlines != 100))
      {
	/* ensure scanlines < 100 ?? */
	/* Kinda redundant on a TV, but maybe not on a VGA box */
	for (x = 0; x < 64; x++)
	  {
	    uint8 r,g,b;
	    
	    r = (uint8)((((NES_palette[x] & 0xFF0000) >> 16) * (scanlines / 100.0))) >> 3;
	    g = (uint8)((((NES_palette[x] & 0xFF00)   >>  8) * (scanlines / 100.0))) >> 2;
	    b = (uint8)((((NES_palette[x] & 0xFF)          ) * (scanlines / 100.0))) >> 3;

	    palette2DC[x] = (r << 11) | (g << 5) | (b);
	  }
      }



  bpp = 16;
  if (magstep == 2) { /*magstep = 2;*/ }
  scanlines = 100;

#ifdef USE_TA
  bytes_per_line = 2 * 512;
	ta_bkg.flags1 = 0x90800000;
	ta_bkg.flags2 = 0x20800440;
	ta_bkg.dummy = 0;
	ta_bkg.x1 = 0.0f;
	ta_bkg.y1 = 480.0f;
	ta_bkg.z1 = 0.2f;
	ta_bkg.argb1 = 0xff000000;
	ta_bkg.x2 = 0.0f;
	ta_bkg.y2 = 0.0f;
	ta_bkg.z3 = 0.2f;
	ta_bkg.argb2 = 0xff000000;
	ta_bkg.x3 = 640.0f;
	ta_bkg.y3 = 480.0f;
	ta_bkg.z3 = 0.2f;
	ta_bkg.argb3 = 0xff000000;

	txraddr = 0x700000; // txr_alloc(512*256*2);
	bufferDC[0] = (uint16 *)ta_texture_map(txraddr) + 8;
	rfb = (char *)bufferDC[0];
	fb = rfb;

#else
  bytes_per_line = 2 * 320 * magstep;
  bufferDC[0] = vram_s;
  bufferDC[1] = vram_s + 320 * magstep * 240 * magstep;
  frameno = 0;
  vid_waitvbl(); 
  vid_set_start((uint32) bufferDC[frameno]); 
  frameno ^= 1; 
  rfb = (char *)(bufferDC[frameno]) + ((magstep == 1 ? 320 : 640) - 256); // temporary kludge to center (see below)
  fb  = rfb;
#endif

  if (verbose)
    {
      fprintf (stderr, "DC: %dbpp, mode %s\n", bpp, magstep == 2 ? "PM_640x480" : "PM_320x240");
    }
	for (x = 0; x < MAX_CONTROLLERS; x++) {
		maple_controller[x] = maple_nth_device(MAPLE_FUNC_CONTROLLER,x);
		fprintf(stderr,"DC: maple_controller[%d] = %02x\n", x, maple_controller[x]);
	}

  fbinit ();
  fprintf(stderr,"DC: fbinit() completed\n");

	init_timer();
  fprintf(stderr,"DC: initdisplay finished\n");
  return 0;
}

/* Update the colors on the screen if the palette changed */
void
UpdateColorsDC(void)
{
  int x;

  fprintf(stderr,"DC: updatecolors begin\n");

  /* Set Background color */
  /* FIXME The commented out stuff should be removed */
//  oldbgcolor = currentbgcolor;
      palette[24] = paletteDC[VRAM[0x3f00] & 63];
//      if (scanlines && (scanlines != 100))
//	palette2[24] = palette2DC[VRAM[0x3f00] & 63];
//     if (oldbgcolor != currentbgcolor /*||currentbgcolor!=bgcolor[currentcache] */ )
//	{
//	  redrawbackground = 1;
//	  needsredraw = 1;
//	}
  /* Set palette tables */
      for (x = 0; x < 24; x++)
	{
	  palette[x] = paletteDC[VRAM[0x3f01 + x + (x / 3)] & 63];
//	  if (scanlines && (scanlines != 100))
//	    palette2[x] = palette2DC[VRAM[0x3f01 + x + (x / 3)] & 63];
	}
  fprintf(stderr,"DC: updatecolors end\n");
}

int
UpdateDisplayDC(void)
{
#ifndef HAVE_DC_TIME
  struct timeval time;
  unsigned int timeframe;
  unsigned static int frame;
#else
  unsigned static int frame = 0;
#endif
  int done;

  static int nodisplay = 0;

  /* do audio update */
  profile_start(2);
  sounder->UpdateAudio();
  profile_end(2);

  /* Check the time.  If we're getting behind, skip a frame to stay in sync. */
#ifndef HAVE_DC_TIME
  gettimeofday (&time, NULL);
  timeframe = (time.tv_sec - renderer_data.basetime) * 50 + time.tv_usec / 20000;     /* PAL */
  timeframe = (time.tv_sec - renderer_data.basetime) * 60 + time.tv_usec / 16666;     /* NTSC */
  frame++;
  if (halfspeed)
    timeframe >>= 1;
  if (doublespeed == 2)
    timeframe <<= 1;
  else if (doublespeed > 2)
    timeframe *= doublespeed;
  if (desync)
    frame = timeframe;
  desync = 0;
  if (frame < timeframe - 20 && frame % 20 == 0)
    desync = 1;                 /* If we're more than 20 frames behind, might as well stop counting. */
#else
  total_frames++;
  if (total_frames % (60*15) == 0) {
	  printf(".");
	}
#endif

/*  fprintf(stderr,"updatedisplay, returning immediately\n"); */
  fprintf(stderr,"updatedisplay begin, frame %d\n", total_frames);
/*  fprintf(stderr,"cksum_add32: %04x\n",cksum_add32(ROM,40976)); */
/*  if (cksum_add32(ROM,40976) != 0x40157a) {
        fprintf(stderr, "cksum error\n");
      	quit();
  } */
  if (! nodisplay)
    {
//	frameskip = 1;
      drawimage (PBL);
      /* Is frameskip placed correctly?  Page flipping is negligible
       * only the VBL wait is timeful. */
      if (!frameskip)
	{
	  UpdateColorsDC();

      vid_waitvbl();
#ifdef USE_TA
{
	poly_hdr_t poly;
	ta_begin_render();
    ta_build_poly_hdr(&poly, TA_OPAQUE, TA_RGB565, 512, 256, txraddr, TA_NO_FILTER);
    ta_commit_poly_hdr(&poly);

	draw_screen_poly();

	ta_commit_eol();

    ta_build_poly_hdr(&poly, TA_TRANSLUCENT, TA_NO_TEXTURE, 0, 0, 0, 0);
	ta_commit_poly_hdr(&poly);

	ta_commit_eol();
	ta_finish_frame();

	rfb = (char *)bufferDC[0];
	// Since we don't use two buffers, we might flicker
	fb = rfb;
}
	
#else
{
	int i;
	uint32 *ptr1 = (uint32 *)bufferDC[frameno] + 12;
	for (i = 0 ; i < 240; i++) {
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  ptr1 += 128;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  ptr1 += 24;
	}
}

	  vid_set_start((uint32) bufferDC[frameno]); 
	  frameno ^= 1;
	  rfb = (char *)(bufferDC[frameno]) + ((magstep == 1 ? 320 : 640) - 256); //temporary kludge to center
          fb =  rfb;
//  sounder->UpdateAudio();

#endif
	  redrawall = needsredraw = 0;
	} else { /* !frameskip */
//      vid_waitvbl();
//  sounder->UpdateAudio();
    }
	}
  else {
      fprintf(stderr,"nodisplay this frame\n");
  }

  frameskip = frameskip + 1;
  if (frameskip > 1) frameskip = 0; /**/

  needsredraw = 0;
  redrawbackground = 0;
  redrawall = 0;

#ifndef HAVE_DC_TIME

  /* Slow down if we're getting ahead */
  if (frame > timeframe + 1 && frameskip == 0)
    {
      usleep (16666 * (frame - timeframe - 1));
    }
#endif

//if (total_frames > 4000) { quit(); }

  /* Input loop */
//  do {
    done = HandleMapleDC();
//  } while (renderer_data.pause_display);  /* Should probably remove this test */
  
#ifndef HAVE_DC_TIME
  /* Check the time.  If we're getting behind, skip next frame to stay in sync. */
  gettimeofday (&time, NULL);
  timeframe = (time.tv_sec - renderer_data.basetime) * 60 + time.tv_usec / 16666;     /* NTSC */
  if (halfspeed)
    timeframe >>= 1;
  if (doublespeed == 2)
    timeframe <<= 1;
  else if (doublespeed > 2)
    timeframe *= doublespeed;
  if (frame >= timeframe || frame % 20 == 0)
    frameskip = 0;
  else
    frameskip = 1;
#endif
  fprintf(stderr,"end updatedisplay\n");
  return done;
}

#ifdef USE_TA
void draw_screen_poly()
{
	vertex_ot_t vert;
	const int x = (640 - (256 * 2)) / 2;   // 64

	vert.flags = TA_VERTEX_NORMAL;
	vert.x = x + 0.0f;
	vert.y = 480.0f;
	vert.z = 1.0f;
	vert.u = 8.0f/512.0f; vert.v = 480.0f/512.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));

	vert.flags = TA_VERTEX_NORMAL;
	vert.y = 0.0f;
    vert.v = 0.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));

	vert.flags = TA_VERTEX_NORMAL;
	vert.x = x + 512.0f;
	vert.y = 480.0f;
	vert.u = 264.0f/512.0f; vert.v = 480.0f/512.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));

	vert.flags = TA_VERTEX_EOL;
	vert.y = 0.0f;
	vert.v = 0.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));
}

#endif

int InitDisplayDCtile(int argc, char **argv)
{
	int x;

	/* ensure hardcode magstep for now */
	magstep = 1;

    vid_init(DM_320x240, PM_RGB565);
    ta_init_defaults();
	
	/* FIXME Clipping needs to be set (to clip overspray pixels) */

	blackpixel = 0x0000;
	whitepixel = 0xffff; /* FIXME whitepixel not used */
	for (x=0; x < 24; x++)
		palette[x] = blackpixel;

	/* Set palette's pixel format */
	*(vuint32 *)PAL_RAM_CTRL = PAL_FMT_ARGB1555;
	/* Set Y output scale to 1.0 -- else everything is blurred */
	*(vuint32 *)SCALER_CFG   = 0x400; 

	/* Turn horizonal clipping on.  We don't have to worry about
	 * manually clipping tiles in the x direction this way. */
	ta_state.pclip_x = ((32 + 256 - 1) << 16) | 32;

	for (x=0; x < 64; x++)
	{
		/* ARGB1555, always opaque */
		uint8 r,g,b;
		r = (NES_palette[x] >> 16 >> 3) & 0x1f;
		g = (NES_palette[x] >>  8 >> 3) & 0x1f;
		b = (NES_palette[x] >>       3) & 0x1f;
		paletteDC[x] = 0x8000 | (r << 10) | (g << 5) | (b);
	}

	tile_cache_initialize();

	bpp = 16; /* FIXME not used */

	if (verbose)
		fprintf (stderr, "DCtile: %dbpp, mode %s\n", bpp, magstep == 2 ? "PM_640x480" : "PM_320x240");

	for (x = 0; x < MAX_CONTROLLERS; x++) {
		maple_controller[x] = maple_nth_device(MAPLE_FUNC_CONTROLLER,x);
		fprintf(stderr,"DCtile: maple_controller[%d] = %02x\n", x, maple_controller[x]);
	}

	fbinit();
	init_timer();

	/* So that we don't time out on the first list */
	pvr_dummy_poly(TA_OPAQUE);
	ta_commit_eol();

	fprintf(stderr,"DCtile: InitDisplay complete\n");
	return 0;
}

void draw_opaque_poly(float x1, float y1, float x2, float y2, float z, float r, float g, float b)
{
	poly_hdr_t poly_opaque;
	vertex_oc_t vert;

	ta_poly_hdr_col(&poly_opaque, TA_OPAQUE);
	ta_commit_poly_hdr(&poly_opaque);

	vert.flags = TA_VERTEX_NORMAL;
	vert.a = 1.0f; vert.r = r; vert.g = g; vert.b = b;
	vert.z = z;
	vert.x = x1; vert.y = y2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x1; vert.y = y1;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y1;
	vert.flags = TA_VERTEX_EOL;
	ta_commit_vertex(&vert, sizeof(vert));
}

void draw_translucent_poly(float x1, float y1, float x2, float y2, float z, float r, float g, float b, float a)
{
	poly_hdr_t poly;
	vertex_oc_t vert;

	ta_poly_hdr_col(&poly, TA_TRANSLUCENT);
	ta_commit_poly_hdr(&poly);

	vert.flags = TA_VERTEX_NORMAL;
	vert.a = a; vert.r = r; vert.g = g; vert.b = b;
	vert.z = z;
	vert.x = x1; vert.y = y2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x1; vert.y = y1;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y1;
	vert.flags = TA_VERTEX_EOL;
	ta_commit_vertex(&vert, sizeof(vert));
}

void draw_textured_pal_poly(uint32 txr_addr, float x, float y, float z, int bank)
{
	poly_hdr_t poly_textured;
	vertex_ot_t vert;
	float x1 = x, y1 = y, x2 = x1 + 8, y2 = y1 + 8;
	float u1 = 0.0f, v1 = 0.0f, u2 = 1.0f, v2 = 1.0f;

	ta_poly_hdr_txr(&poly_textured, TA_TRANSLUCENT, TA_PAL4BPP, 8, 8, txr_addr, TA_NO_FILTER);
	poly_textured.flags4 &= ~PAL_BANK_MASK_4BPP;
	poly_textured.flags4 |= bank << PAL_BANK_SHIFT_4BPP;
	ta_commit_poly_hdr(&poly_textured);

	vert.flags = TA_VERTEX_NORMAL;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	vert.oa = 1.0f; vert.or = 1.0f; vert.og = 1.0f; vert.ob = 1.0f;
	vert.z = z;
	vert.x = x1; vert.y = y2;
	vert.u = u1; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x1; vert.y = y1;
	vert.u = u1; vert.v = v1;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y2;
	vert.u = u2; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y1;
	vert.u = u2; vert.v = v1;
	vert.flags = TA_VERTEX_EOL;
	ta_commit_vertex(&vert, sizeof(vert));
}

void draw_textured_pal_poly_flip(uint32 txr_addr, float x, float y, float z, int hflip, int vflip, int bank)
{
	poly_hdr_t poly_textured;
	vertex_ot_t vert;
	float x1 = x, y1 = y, x2 = x1 + 8, y2 = y1 + 8;
	float u1 = 0.0f, v1 = 0.0f, u2 = 1.0f, v2 = 1.0f;

	if (hflip) {u1 = 1.0f; u2 = 0.0f; }
	if (vflip) {v1 = 1.0f; v2 = 0.0f; }

	ta_poly_hdr_txr(&poly_textured, TA_TRANSLUCENT, TA_PAL4BPP, 8, 8, txr_addr, TA_NO_FILTER);
	poly_textured.flags4 &= ~PAL_BANK_MASK_4BPP;
	poly_textured.flags4 |= bank << PAL_BANK_SHIFT_4BPP;
	ta_commit_poly_hdr(&poly_textured);

	vert.flags = TA_VERTEX_NORMAL;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	vert.oa = 1.0f; vert.or = 1.0f; vert.og = 1.0f; vert.ob = 1.0f;
	vert.z = z;
	vert.x = x1; vert.y = y2;
	vert.u = u1; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x1; vert.y = y1;
	vert.u = u1; vert.v = v1;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y2;
	vert.u = u2; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y1;
	vert.u = u2; vert.v = v1;
	vert.flags = TA_VERTEX_EOL;
	ta_commit_vertex(&vert, sizeof(vert));
}

void draw_textured_pal_poly_partial(uint32 txr_addr, float x, float y, float z, int clip_top, int clip_bot, int bank)
{
	poly_hdr_t poly_textured;
	vertex_ot_t vert;
	float x1 = x, y1 = y, x2 = x1 + 8, y2 = y1 + clip_bot - clip_top;
	float u1 = 0.0f, v1 = clip_top / 8.0f, u2 = 1.0f, v2 = clip_bot / 8.0f;

	ta_poly_hdr_txr(&poly_textured, TA_TRANSLUCENT, TA_PAL4BPP, 8, 8, txr_addr, TA_NO_FILTER);
	poly_textured.flags4 &= ~PAL_BANK_MASK_4BPP;
	poly_textured.flags4 |= bank << PAL_BANK_SHIFT_4BPP;
	ta_commit_poly_hdr(&poly_textured);

	vert.flags = TA_VERTEX_NORMAL;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	vert.oa = 1.0f; vert.or = 1.0f; vert.og = 1.0f; vert.ob = 1.0f;
	vert.z = z;
	vert.x = x1; vert.y = y2;
	vert.u = u1; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x1; vert.y = y1;
	vert.u = u1; vert.v = v1;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y2;
	vert.u = u2; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y1;
	vert.u = u2; vert.v = v1;
	vert.flags = TA_VERTEX_EOL;
	ta_commit_vertex(&vert, sizeof(vert));
}

int UpdateDisplayDCtile(void)
{
	/* The opaque background plane should go -after- the ta_finish_frame,
	 * because drawimage_tile will feed the TA translucent polygons and
	 * we can't mix opaque and translucent--unless we make the background plane
	 * "translucent" (with opaque alpha) too. */
	/* Alternatively, we may be able to enable the TA's real background plane. */

	float r, g, b;


//	pvr_dummy_poly(TA_OPAQUE);
	pvr_dummy_poly(TA_TRANSLUCENT);
	drawimage(PBL);
	ta_commit_eol();
	ta_finish_frame();

	sounder->UpdateAudio();
	UpdateColorsDCtile();

	total_frames++;
	if (total_frames % (60*15) == 0) printf(".");

	/* This is a very lossy computation, isn't it. */
	r = (palette[24] >> 10) & 0x1f;
	g = (palette[24] >>  5) & 0x1f;
	b = (palette[24]      ) & 0x1f;
	r = r / 31.0f; g = g / 31.0f; b = b / 31.0f; /* 32? */

	draw_opaque_poly(32, 0, 256 + 32, 240, 100, r, g, b);

//	pvr_dummy_poly(TA_OPAQUE);
	ta_commit_eol();

	return HandleMapleDC();
}

static unsigned int (*palette_dc)[64][16] = (unsigned int (*)[64][16])0xa05f9000;

void UpdateColorsDCtile(void)
{
	int i;

	/* Use palette[24] to hold background color. */
	palette[24] = paletteDC[VRAM[0x3f00] & 63];

	/* Unrolled into two loops for clarity. */
	/* Tile palette; multiples of 4 are transparent. */
	for (i = 0; i < 16; i++)
		(*palette_dc)[PAL_BANK_TILE][i] = (i & 3) ? paletteDC[VRAM[0x3f00 + i] & 63] : 0x0000;

	/* Sprite palette; multiples of 4 are transparent. */
	for (i = 0; i < 16; i++)
		(*palette_dc)[PAL_BANK_SPRITE][i] = (i & 3) ? paletteDC[VRAM[0x3f10 + i] & 63] : 0x0000;
}


#endif
