/*
 * This file is part of the TuxNES project codebase.
 *
 * Please see the README and COPYING files for more information regarding
 * this project.
 *
 * $Id: fb.c,v 1.5 2002/02/28 00:28:48 jim Exp $
 *
 * Description: Framebuffer/pixmap rendering; the main rendering
 * function drawimage* is included (several times) from pixels.h.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef HAVE_DC
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_FEATURES_H
#include <features.h>
#endif
#else
#include "dc-glue.h"
#undef printf
#endif

#include "globals.h"
#include "mapper.h"
#include "renderer.h"

/* globals */

unsigned int    nextline;
unsigned int	bpp;
unsigned int    bpu;
unsigned int    bitmap_pad;
unsigned int	depth;
unsigned int    bytes_per_line;
unsigned int	lsb_first;
unsigned int	lsn_first; /* nybbles swapped? Ick! */
unsigned int    pix_swab;
unsigned int	vline = 0;
unsigned int	vscan = 0;
unsigned int	frameskip = 0;
unsigned int	vwrap = 0;
unsigned int    scanpage = 0;
unsigned int    scanlines = 100;
/* These should be moved here (from x11.c) even in non-DC case */
#ifdef HAVE_DC
int             magstep = 0;
int	indexedcolor = 1;
unsigned int	currentbgcolor, oldbgcolor;
unsigned char	needsredraw = 1;	/* Refresh screen display */
unsigned char	redrawbackground = 1;	/* Redraw tile background */
unsigned char	redrawall = 1;		/* Redraw all scanlines */
#endif /* HAVE_DC */
char	*fb = 0;
char	*rfb = 0;
int	 palette_alloc[25];
int	*palette = palette_alloc;
int	 palette_alloc2[25];
int	*palette2 = palette_alloc2;

void	(*drawimage)(int);
void	fbinit(void);

void	mmc2_4_latch(int);
void	mmc2_4_latchspr(int);

#undef DOUBLE
#define DOUBLE 0

#undef SCANLINES
#define SCANLINES 0

#define BPP 16
#include "pixels.h"

#undef BPP

extern void drawimage_tile(int);

void
drawimage_old(int endclock)
{
  return;
}

void
fbinit(void)
{
	/* this is the default inter-line skip */
	nextline = bytes_per_line;
	if ((renderer->_flags & RENDERER_OLD) ||
			(renderer->InitDisplay == InitDisplayNone))
	{
		/* Point drawimage to the no-op version */
		drawimage = drawimage_old;
	}
	else if (renderer->_flags & RENDERER_TILE) {
		drawimage = drawimage_tile;
		nextline = 0;  /* nextline is not used */
	}
	else if (bpp == 16)
	{
		drawimage = (magstep == 2) ? (scanlines != 100) ? drawimage_old : drawimage_old : &drawimage16;
		nextline = bytes_per_line / 2;
	}
	else
	{
		fprintf (stderr, "Don't know how to handle %dbpp\n", bpp);
		exit (1);
	}
}
