/*
 * This file is part of the TuxNES project codebase.
 *
 * Please see the README and COPYING files for more information regarding
 * this project.
 *
 * $Id: pixels.h,v 1.13 2002/02/28 00:28:48 jim Exp $
 *
 * Description: This file is included several times with different bpps
 * and enlargement/scanline combinations defined. See fb.c.
 */

#include "consts.h"
#include "dc-glue.h"

#undef pixel_t

#undef endian_fix

#if (BPP==16)
#define endian_fix(x) (x)
#define pixel_t unsigned short int
void	drawimage16(int);
#if DOUBLE
#  error DOUBLE not supported in 16bpp
#elif SCANLINES
#  error SCANLINES not supported in 16bpp
#endif
#else
#  error Only 16bpp is supported.
#endif

#undef rpixmap
#define rpixmap ((pixel_t *)rfb)

#undef pixmap
#define pixmap ((pixel_t *)fb)

void
drawimage16(int endclock)
{
  static unsigned int lastclock = 0;
  static unsigned int curhscroll = 0;
  unsigned int curclock = lastclock;
  unsigned int baseaddr = ((RAM[0x2000] & 0x10) << 8);  /* 0 or 0x1000 */
  unsigned int currentline = lastclock / HCYCLES;
  unsigned int hposition = lastclock % HCYCLES;
  unsigned int x;
  int s;
  int hflip, vflip, behind;
  int d1, d2;
  int spritebase = (RAM[0x2000] & 0x08) << 9;   /* 0x0 or 0x1000 */
  int spritesize = 8 << ((RAM[0x2000] & 0x20) >> 5);    /* 8 or 16 */
  int spritetile;
  unsigned char linebuffer[256];
  unsigned char bgmask[256+8];  // Allow for maximum 8 extra overspray pixels
  unsigned char *bgmask_ptr;
  /* tile, byte1/2 don't need to be static. */
  static unsigned int tile;
  static int bit;
  static unsigned char byte1, byte2;
  static pixel_t curpal[4];
  static unsigned char tilecolor;
  static pixel_t *rptr, *rptr0;
  static pixel_t *ptr, *ptr0;

  if (frameskip)
    {
/*	  fprintf(stderr,"drawimage: frameskip\n");  */
      return;
    }
	  // eprintf("drawimage: no frameskip\n"); 
  // eprintf("cur %d end %d last %d ptr %p\n", curclock, endclock, lastclock, ptr);

/* 113 2/3 cpu cycles per scanline */
/* = 341 ppu cycles per scanline */
/* = 81840 ppu cycles per frame (not counting vblank) */
/* = 89342 ppu cycles per frame (including vblank) */
/*
  if (endclock == PBL) {
  	lastclock = 0;
	return;
  }
  */

  if (endclock > PBL)
    return;
  if (curclock > PBL)
    curclock = 0;
  if (endclock <= curclock)
    return;

  profile_start(0);

  profile_start(4);

  if (curclock == 0)
    {
      /* Begin new frame */
      vline = vscrollreg >> 3;
      vscan = vscrollreg & 7;
      vwrap = 0;
      if (osmirror)
        scanpage = 0x2000;
      else if (nomirror)
        scanpage = 0x2000 + ((RAM[0x2000] & 3) << 10);
      else if (hvmirror == 0)
        scanpage = 0x2000 + ((RAM[0x2000] & 1) << 10);  /* v-mirror, h-layout */
      else if (hvmirror == 1)
        scanpage = 0x2000 + ((RAM[0x2000] & 2) << 9);   /* h-mirror, v-layout */
      curpal[0] = endian_fix(palette[24]);
      rptr0 = rptr = rpixmap;
      ptr0 = ptr = pixmap;
    }

{
	int endclock_hpos, endclock_hpos0;
	endclock_hpos = endclock % HCYCLES;
	endclock_hpos0 = endclock - endclock_hpos;
	if (endclock_hpos < 85) {
		// Ended in hblank; increase to end of hblank.
		endclock = endclock_hpos0 + 84;
	} else {
		// Ended mid-line; increase to end of next hblank.
		endclock = endclock_hpos0 + HCYCLES + 84;
	}
}

  if (hposition >= 85)
    {
      /* In scanline */
	  // Poor man's assert
	  *((volatile unsigned int *)3);
      x = hposition - 85 + curhscroll;
    }
  else
    {
      /* In hblank */
		/* FIXME we can eliminate curhscroll variable now */
      curhscroll = hscrollreg;
      x = curhscroll;
      if (osmirror)
         scanpage = 0x2000;
      else if (nomirror)
         scanpage = 0x2000 + (((RAM[0x2000] & 3) << 10) ^ (vwrap << 11));
      else if (hvmirror == 0)
         scanpage = 0x2000 + ((RAM[0x2000] & 1) << 10);      /* v-mirror, h-layout */
      else if (hvmirror == 1)
         scanpage = 0x2000 + (((RAM[0x2000] & 2) << 9) ^ (vwrap << 10));     /* h-mirror, v-layout */
      tile = VRAM[scanpage + ((x & 255) >> 3) + (vline << 5)];
      mmc2_4_latch(baseaddr + (tile << 4) + vscan);
      mmc2_4_latch(baseaddr + (tile << 4) + vscan + 8);
      byte1 = VRAM[baseaddr + (tile << 4) + vscan];
      byte2 = VRAM[baseaddr + (tile << 4) + vscan + 8];
      bit = (~x) & 7;
      curclock += 85 - hposition;
	  /* hposition is set in init code, that seems redundant */
      hposition = 85;
      tilecolor = VRAM[scanpage + 0x3C0 + ((x & 255) >> 5) + ((vline & 28)
                                                              << 1)] >>
        ((vline
          &
          2)
         << 1);
      if (x & 16)
        tilecolor >>= 2;
      curpal[1] = endian_fix(palette[3 * (tilecolor & 3)]);
      curpal[2] = endian_fix(palette[3 * (tilecolor & 3) + 1]);
      curpal[3] = endian_fix(palette[3 * (tilecolor & 3) + 2]);
    }

	profile_end(4);

/* tight loop */
#if 0
	while (curclock < endclock)
	{
		while (hposition < HCYCLES && curclock < endclock)
		{
			curclock++;
			hposition++;
		} 
		if (hposition >= HCYCLES) {
			hposition=85;
			curclock+=85;
		}
	}
#else

	profile_start(5);
	while (curclock < endclock)
	{

	{
	     int t1, t2, t3;
	     t1 = scanpage + (vline << 5);
	     t2 = baseaddr + vscan;
	     t3 = t2 + 8;

		/* Force x and bit to be 8-pixel (tile) aligned.
		   Shift image x&7 pixels to the left (temporarily,
		   until we can just skip the first x&7 pixels.) */
		/* Also, adjust hposition and curclock to display
		   an extra x&7 pixels from the next tile on the right. */
	    ptr -= (x&7);
		hposition -= (x&7);
		x &= ~7;
		bgmask_ptr = &bgmask[hposition - 85];

			profile_start(9);
		while (hposition < HCYCLES) {
			profile_start(7);
          if ((RAM[0x2001] & 8))
            {
//				bit = 7;
#if 0
#   if 0
			 	*(unsigned int *)ptr = (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> bit) & 0x01) | ( ((byte2 >> (bit-1)) & 0x02) ) )]) |
                                   (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> (bit-1)) & 0x01) | ( ((byte2 >> (bit-2)) & 0x02) ) )] << 16);
			  ptr += 2;
              bit-=2;
			 	*(unsigned int *)ptr = (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> bit) & 0x01) | ( ((byte2 >> (bit-1)) & 0x02) ) )]) |
                                   (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> (bit-1)) & 0x01) | ( ((byte2 >> (bit-2)) & 0x02) ) )] << 16);
			  ptr += 2;
              bit-=2;
			 	*(unsigned int *)ptr = (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> bit) & 0x01) | ( ((byte2 >> (bit-1)) & 0x02) ) )]) |
                                   (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> (bit-1)) & 0x01) | ( ((byte2 >> (bit-2)) & 0x02) ) )] << 16);
			  ptr += 2;
              bit-=2;
			 	*(unsigned int *)ptr = (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> bit) & 0x01) | ( ((byte2 >> (bit-1)) & 0x02) ) )]) |
                                   (curpal[/* bgmask[hposition - 85] = */ ( ((byte1 >> (bit-1)) & 0x01) | ( ((byte2 >> (bit-2)) & 0x02) ) )] << 16);
			  ptr += 2;
              bit-=2;
#   endif
			 	*ptr = curpal[bgmask[hposition - 85] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
				*ptr = curpal[bgmask[hposition - 85 + 1] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
			 	*ptr = curpal[bgmask[hposition - 85 + 2] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
			 	*ptr = curpal[bgmask[hposition - 85 + 3] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
#define SIXTEENBITWRITES
#  ifdef SIXTEENBITWRITES
#if 1
			 	*ptr = curpal[bgmask[hposition - 85 + 4] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
			 	*ptr = curpal[bgmask[hposition - 85 + 5] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
			 	*ptr = curpal[bgmask[hposition - 85 + 6] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
			 	*ptr = curpal[bgmask[hposition - 85 + 7] = 
				         ( ((byte1 >> bit) & 0x01) | 
					   ( ((byte2 >> (bit-1)) & 0x02) ) )];
				ptr++;
				bit--;
#endif
#  else
			   /* fake 32-bit writes by skipping half the pixels */
			   /* A 32-bit write has to be 32-bit aligned, so we
			      can't scroll correctly in the current implementation.  */
				ptr+=4;
				bit -=4;
#  endif
#else
				// must be easier way to do this
				// like repeated rcl or somesuch
				// this is substantially slower for some reason
				// probably compiler optimizes differently
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1 >> 7) & 0x01) | 
					   ( ((byte2 >> 6) & 0x02) ) )];
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1 >> 6) & 0x01) | 
					   ( ((byte2 >> 5) & 0x02) ) )];
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1 >> 5) & 0x01) | 
					   ( ((byte2 >> 4) & 0x02) ) )];
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1 >> 4) & 0x01) | 
					   ( ((byte2 >> 3) & 0x02) ) )];
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1 >> 3) & 0x01) | 
					   ( ((byte2 >> 2) & 0x02) ) )];
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1 >> 2) & 0x01) | 
					   ( ((byte2 >> 1) & 0x02) ) )];
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1 >> 1) & 0x01) | 
					   ( ((byte2) & 0x02) ) )];
			 	*ptr++ = curpal[*bgmask_ptr++ =
				         ( ((byte1) & 0x01) | 
					   ( ((byte2 <<1) & 0x02) ) )];
				
#endif
            }
			else {
				*ptr++ = *curpal;
				*ptr++ = *curpal;
				*ptr++ = *curpal;
				*ptr++ = *curpal;
				*ptr++ = *curpal;
				*ptr++ = *curpal;
				*ptr++ = *curpal;
				*ptr++ = *curpal;
			}
          x+=8;
          hposition+=8;
		  // This needs to be per-pixel.
		  
          if (x == 256) {
          	if ((!osmirror) && (nomirror || !hvmirror))
		      { 
					scanpage ^= 0x400;        // bit 8 of x -> bit 10 of addr
				  t1 = scanpage + (vline << 5);
			  } 
			  x = 0;
		  }

#if 0
			  
              tile = VRAM[scanpage + ((x & 255) >> 3) + (vline << 5)];
//	      mmc2_4_latch(baseaddr + (tile << 4) + vscan);
//	      mmc2_4_latch(baseaddr + (tile << 4) + vscan + 8);
              byte1 = VRAM[baseaddr + (tile << 4) + vscan];
              byte2 = VRAM[baseaddr + (tile << 4) + vscan + 8]; 
#else

			/* It's not conclusive whether or not this precalculation helps us. */
              tile = VRAM[t1 + (x >> 3)];
//		      mmc2_4_latch(t2 + (tile << 4));
//			mmc2_4_latch(t3 + (tile << 4));
              byte1 = VRAM[t2 + (tile << 4)];
              byte2 = VRAM[t3 + (tile << 4)];

#endif

//              bit = 7;
              if ((x & 0xf) == 0)
                {
                  if ((x & 0x1f) == 0)
                    tilecolor = VRAM[scanpage + 0x3C0 + (x >> 5) + ((vline & 28) << 1)] >> ((vline & 2) << 1);
                  else
                    tilecolor >>= 2;
                  curpal[1] = endian_fix(palette[3 * (tilecolor & 3)]);
                  curpal[2] = endian_fix(palette[3 * (tilecolor & 3) + 1]);
                  curpal[3] = endian_fix(palette[3 * (tilecolor & 3) + 2]);
                }
			profile_end(7);
        }
		profile_end(9);
	} // end namespace

	profile_start(1);

          if ((RAM[0x2001] & 16) && 1) 
            {
              /* Draw sprites */
#ifdef MMC2_LATCHSPR
              for (s = 0; s < 64; s++) {
                if (spriteram[s * 4] < currentline && spriteram[s * 4] < 240 && spriteram[s * 4 + 3] < 249 
                      && (spriteram[s * 4] + spritesize >= currentline))
                {
                  spritetile = spriteram[s * 4 + 1];
                  if ((spritetile == 0xfd) || (spritetile == 0xfe)) {
                    mmc2_4_latchspr(spritetile);
                  }
                }
              }
#endif
              for (s = 63*4; s >= 0; s-=4)
                {
					/* is currentline ever >= 240? */
                  if (spriteram[s] < currentline && spriteram[s] < 240 && spriteram[s + 3] < 256)
                    {
                      if (spriteram[s] + spritesize >= currentline)
                        {
                          spritetile = spriteram[s + 1];
                          if (spritesize == 16)
                            spritebase = (spritetile & 1) << 12;
                          behind = spriteram[s + 2] & 0x20;
                          hflip = spriteram[s + 2] & 0x40;
                          vflip = spriteram[s + 2] & 0x80;

                          /* This finds the memory location of the tiles, taking into account
                             that vertically flipped sprites are in reverse order. */
                          if (vflip)
                            {
                              if (spriteram[s] >= ((signed int) currentline) - 8)       /* 8x8 sprites and first half of 8x16 sprites */
                                {
                                  d1 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + spritesize * 2 - 8 - currentline + spriteram[s]];
                                  d2 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + spritesize * 2 -     currentline + spriteram[s]];
                                }
                              else
                                /* Do second half of 8x16 sprites */
                                {
                                  d1 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + spritesize * 2 - 16 - currentline + spriteram[s]];
                                  d2 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + spritesize * 2 -  8 - currentline + spriteram[s]];
                                }
                            }
                          else
                            {
                              if (spriteram[s] >= ((signed int)
                                                        currentline) - 8)       /* 8x8 sprites and first half of 8x16 sprites */
                                {
                                  d1 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + currentline - 1 - spriteram[s]];
                                  d2 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + currentline + 7 - spriteram[s]];
                                }
                              else
                                /* Do second half of 8x16 sprites */
                                {
                                  d1 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + currentline +  7 - spriteram[s]];
                                  d2 = VRAM[spritebase + ((spritetile & (~(spritesize >> 4))) << 4) + currentline + 15 - spriteram[s]];
                                }
                            }
                          for (x = 7 * (!hflip); x < 8 && x >= 0; x += 1 - ((!hflip) << 1))
                            {
                              if (spriteram[s + 3] + x < 256) {
                                // Original code wrote pixel twice if behind, second time set to zero.  We just skip it.
                                if (! (behind && (d1 | d2) && bgmask[spriteram[s + 3] + x]))
                                {
                                  if (d1 & d2 & 1)
                                    ptr0[spriteram[s + 3] + x] = palette[12 + 2 + ((spriteram[s + 2] & 3) * 3)];
                                  else if (d1 & 1)
                                    ptr0[spriteram[s + 3] + x] = palette[12 +     ((spriteram[s + 2] & 3) * 3)];
                                  else if (d2 & 1)
                                    ptr0[spriteram[s + 3] + x] = palette[12 + 1 + ((spriteram[s + 2] & 3) * 3)];
                                }
                              }
                              d1 >>= 1;
                              d2 >>= 1;
                            }
                        }
                    }
                }

            }

		profile_end(1);
		profile_start(8);

          /* Next line */
	  // LOOKHERE -- remove rptr
//	  rptr = rptr0 + nextline;
	  ptr = ptr0 + nextline;
	  // LOOKHERE -- remove rptr
//	  rptr0 = rptr;
	  ptr0 = ptr;

          currentline++;
          curhscroll = hscrollreg;
          x = curhscroll;
          vscan++;
          if (vscan >= 8)
            {
              vscan = 0;
              vline++;
              vline &= 31;
              if (vline == 30)
                {
                  vline = 0;
                  vwrap ^= 1;
                }
            }
          if (osmirror)
            scanpage = 0x2000;
          else if (nomirror)
            scanpage = 0x2000 + (((RAM[0x2000] & 3) << 10) ^ (vwrap << 11));
          else if (hvmirror == 0)
            scanpage = 0x2000 + ((RAM[0x2000] & 1) << 10);      /* v-mirror, h-layout */
          else if (hvmirror == 1)
            scanpage = 0x2000 + (((RAM[0x2000] & 2) << 9) ^ (vwrap << 10));     /* h-mirror, v-layout */
          tile = VRAM[scanpage + ((x & 255) >> 3) + (vline << 5)];
//	  mmc2_4_latch(baseaddr + (tile << 4) + vscan);
//	  mmc2_4_latch(baseaddr + (tile << 4) + vscan + 8);
          byte1 = VRAM[baseaddr + (tile << 4) + vscan];
          byte2 = VRAM[baseaddr + (tile << 4) + vscan + 8];
//          bit = (~x) & 7;
          tilecolor = VRAM[scanpage + 0x3C0 + ((x & 255) >> 5) + ((vline &
                                                                   28) << 1)]
            >> ((vline & 2) << 1);
          if (x & 16)
            tilecolor >>= 2;
          curpal[1] = endian_fix(palette[3 * (tilecolor & 3)]);
          curpal[2] = endian_fix(palette[3 * (tilecolor & 3) + 1]);
          curpal[3] = endian_fix(palette[3 * (tilecolor & 3) + 2]);

		curclock += HCYCLES;
          hposition = 85;
		  profile_end(8);
    }
#endif
	profile_end(5);

  // Redundant >, we exit above if endclock > PBL.
  // This means "if endclock == PBL".
  if ((lastclock = endclock) >= PBL) {
    lastclock = 0;
//  eprintf(" exit cur %d end %d last %d ptr %p\n", curclock, endclock, lastclock, ptr);
  }
 	profile_end(0); 
}
