/*
 * This file is part of the TuxNES project codebase.
 *
 * Please see the README and COPYING files for more information regarding
 * this project.
 *
 * $Id: sound.h,v 1.3 2002/01/12 02:29:51 jim Exp $
 *
 * Description: sound interface
 */

/* exports */

extern void		SoundEvent(long addr, unsigned char value);
extern char     	SoundGetLengthReg(void);
extern void		InitSoundParams(int argc);
extern void		ProgressiveSoundUpdate(int num_samples);

extern int		InitAudioDC(int argc, char **argv);
extern void		UpdateAudioDC(void);
extern int		ShutdownAudioDC(void);
extern int		InitAudioMute(int argc, char **argv);
extern void		UpdateAudioMute(void);
extern int		ShutdownAudioMute(void);
extern int		InitAudioNone(int argc, char **argv);
extern void		UpdateAudioNone(void);
extern int		ShutdownAudioNone(void);

struct SampleFormat {
     char *name, *fullname;
     int number;
};

/* the currently selected sample format */
extern struct SampleFormat *sample_format;

/* table of sample formats, terminated by { 0, 0, 0, 0 } */
extern struct SampleFormat sample_formats[];

/* lookup table for 12bit->Mu-Law sample conversion */
extern unsigned char ulaw [0x4000];

/* global sound parameters */
extern struct SoundConfig {
     char	*audiofile;
     int	audiorate;
     int	audiostereo;
     float	max_sound_delay;
     int	reverb;
} sound_config;

struct Sounder {
	char *name, *fullname;
	int (*InitAudio)(int argc, char **argv);
	void (*UpdateAudio)(void);
	int (*ShutdownAudio)(void);
};

extern struct Sounder *sounder;
extern struct Sounder sounders[];

extern int bytes_per_sample;
extern int sample_format_number;
extern int signed_samples;
extern unsigned char *audio_buffer;
extern unsigned int  samples_per_vsync;

/* sample format numbers */
#ifndef AFMT_MU_LAW
#  define AFMT_MU_LAW 1
#endif
#ifndef AFMT_U8
#  define AFMT_U8 8
#endif
#ifndef AFMT_S16_LE
#  define AFMT_S16_LE 16
#endif
#ifndef AFMT_S16_BE
#  define AFMT_S16_BE 32
#endif
#ifndef AFMT_S8
#  define AFMT_S8 64
#endif
#ifndef AFMT_U16_LE
#  define AFMT_U16_LE 128
#endif
#ifndef AFMT_U16_BE
#  define AFMT_U16_BE 256
#endif

#define UPDATE_FREQ 60.0
